/*	Additional BN module,
	Used in Hariyanto Lim custom RSA 
	Created by Hariyanto Lim, last compiled Sunday, 22nd October 2001 */

#include <stdio.h>
#include <bn.h>

int BN_XOR(BIGNUM *target, BIGNUM *source1, BIGNUM *source2)
{
	BIGNUM *temp;
	BN_ULONG i, source1top, source2top;
	BN_ULONG *target_d, *source2_d;

	// find whether target is actually source1
	if(target != source1)
	{	// not equal, so make the content equal
		if(!BN_copy(target, source1))
			return(-1); // copy error
	}

	source1top = target->top;
	source2top = source2->top;

	// make sure source1 is bigger
	if(source1top < source2top)
	{	// source2 is bigger, so swap them
		temp = target;
		target = source2;
		source2 = temp;

		source1top = target->top;
		source2top = source2->top;
	}
	
	// the main XOR function
	target_d = target->d;
	source2_d = source2->d;
	for(i = 0; i < source2top; i++)
	{
		target_d[i] ^= source2_d[i];
	}

	return(0);
}

BIGNUM *BN_AND(BIGNUM *source1, BIGNUM *source2)
{	// this function designed this way because we prevent
	// if source1 = target or source2 = target
	BIGNUM *target = BN_new();
	BN_ULONG i, length;
	BN_ULONG *target_d, *source1_d;

	// copy target = source2 ... because we will only operate on target only ..
	if(!BN_copy(target, source2))
		return(NULL); // copy error
	
	// get total loop
	if(source1->top > target->top)
		length = target->top;
	else
		length = source1->top;	
	
	// the main AND function ... target &= source1 
	// the result is target <= source1
	target_d = target->d;
	source1_d = source1->d;
	for(i = 0; i < length; i++)
	{
		target_d[i] &= source1_d[i];
	}

	return(target);
}