/*	rsa_lib/rsa_pem.c
	This is PEM (privacy-Enhanced Mail) related code
	read these two documents below for more information
	http://www.faqs.org/rfcs/rfc1421.html
	http://www.faqs.org/rfcs/rfc1521.html		
	Last compiled Tuesday, 13th November 2001 */

/* Written by Hariyanto Lim 
 * hmaxf_urlcr@yahoo.com
 * www.geocities.com/hmaxf_urlcr
 * ====================================================================
 * Copyright (c) 2001.  HMaxF Project, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Copyright remains to Hariyanto Lim, and as such any Copyright notices in
 *	  the code are not to be removed. If this package is used in a product, 
 *    Hariyanto Lim should be given attribution as the author of the parts of the library used.
 *    This can be in the form of a textual message at program startup or
 *    in documentation (online or textual) provided with the package.
 * ===================================================================
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
 * IN NO EVENT SHALL I, HARIYANTO LIM, THE AUTHOR BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include <stdio.h>
#include <malloc.h> // malloc(), free()
#include <memory.h> // memcpy()
#include <rsa.h> // to get the definition of buffer size
const unsigned char PEM_encode_data[] = {	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',	//0-7
											'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',	//8-15
											'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',	//16-23
											'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', //24-31
											'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', //32-39
											'o', 'p', 'q', 'r', 's', 't', 'u', 'v', //40-47
											'w', 'x', 'y', 'z', '0', '1', '2', '3', //48-55
											'4', '5', '6', '7', '8', '9', '+', '/', }; //56-63

/*	Since:
	(1)	PEM data is not sorted and
	(2)	the nature of C/C++ array definition
		which have to start from 0(zero) 
		we have to fill the gap in between with ascii 0 = NULL .. 
		which also serve as error control */
const unsigned char PEM_decode_data[] = {0,  0,  0,  0,  0,  0,  0,  0, //0-7
										 0,  0,  0,  0,  0,  0,  0,  0,	//8-15
										 0,  0,  0,  0,  0,  0,  0,  0,	//16-23
										 0,  0,  0,  0,  0,  0,  0,  0, //24-31
										 0,  0,  0,  0,  0,  0,  0,  0,	//32-39
										 0,  0,  0, 62,  0,  0,  0, 63,	//40-47
										52, 53, 54, 55, 56, 57, 58, 59,	//48-55
										60, 61,  0,  0,  0,  0,  0,  0,	//56-63 ... 61 = '='
										 0,  0,  1,  2,  3,  4,  5,  6, //64-71
										 7,  8,  9, 10, 11, 12, 13, 14, //72-79
										15, 16, 17, 18, 19, 20, 21, 22, //80-87
										23, 24, 25,  0,  0,  0,  0,  0, //88-95
										 0, 26, 27, 28, 29, 30, 31, 32, //96-103
										33, 34, 35, 36, 37, 38, 39, 40, //104-111
										41, 42, 43, 44, 45, 46, 47, 48, //112-119
										49, 50, 51, }; //120-122

const unsigned char pad = '=';
const unsigned char CRLF[] = { 0x0D, 0x0A }; // CR = 13, LF = 10 
const unsigned long CRLF_length = 2;
const unsigned long max_total_character_in_a_line = 64;
const unsigned long max_total_character_in_a_line_plus_CRLF = 66; //max_total_character_in_a_line + CRLF_length;
const unsigned long small_size = 3, big_size = 4, six = 6;

//	slow but works ... target SHOULD be at least 33.3% bigger than source, 
//	but I recommend to use 100% more than source, just in case it is overlapping
//	the return value is the real size required
unsigned long PEM_encode(unsigned char *source, long source_size, unsigned char *result)
{	// convert 8 bits to 6 bits .. increase 33.3% in size	
	const unsigned long max_six_bits = 63;
	unsigned long total_block = source_size / small_size;
	unsigned long left_over = source_size % small_size;
	unsigned long i, y;
	unsigned long *source_pointer_long = NULL;
	unsigned char *source_pointer = NULL;
	unsigned long target_size, final_size, current_value;
	unsigned long total_line;
	unsigned char *target = result;
	unsigned char original_buffer_pointer[buffer_size]; // temporary buffer ... use array to avoid buggy malloc 
	unsigned char *buffer = original_buffer_pointer;
	
	// allocate space
	if(left_over)
		total_block += 1;

	target_size = total_block * big_size; // actual target size
	
	total_line = target_size / max_total_character_in_a_line; // find total line
		
	final_size = target_size + (total_line * CRLF_length); // calculate return value, REAL target size

	// convert 24bits of 8bits to 32bits of 6bits	
	source_pointer = source;	
	for(i = 1; i < total_block; i++) // start from 1 to avoid subtraction by one to total_block ... optimize
	{		
		source_pointer_long = (unsigned long *)source_pointer;
		current_value = *source_pointer_long;

		// store them		
		for(y = 0; y < big_size; y++)
		{
			*(target + y) = PEM_encode_data[current_value & max_six_bits]; // direct convert to PEM
			current_value >>= six;
		}

		// increase counter
		source_pointer += small_size;
		target += big_size;
	}

	// handle the left over if any
	if(left_over)
	{		
		current_value = 0; // reset
		memcpy(&current_value, source_pointer, left_over);

		if(left_over == 1)
		{
			*(target) = PEM_encode_data[current_value & max_six_bits];
			current_value >>= six;
			*(target + 1) = PEM_encode_data[current_value];
			*(target + 2) = pad;
			*(target + 3) = pad;
		}
		else
		{
			*(target) = PEM_encode_data[current_value & max_six_bits];
			current_value >>= six;
			*(target + 1) = PEM_encode_data[current_value & max_six_bits];
			current_value >>= six;
			*(target + 2) = PEM_encode_data[current_value];
			*(target + 3) = pad;
		}		
	}
	
	// rewind the target pointer
	target = result;

	// now translate the six bit value to the standard PEM with CRLF character after 64 character 
	for(i = 0; i < total_line; i++)
	{
		memcpy(buffer, target, max_total_character_in_a_line);
				
		memcpy(buffer + max_total_character_in_a_line, CRLF, CRLF_length); // add CRLF ... 
		
		// increase counter
		target += max_total_character_in_a_line;
		buffer += max_total_character_in_a_line_plus_CRLF;
	}

	// handle the left over if any
	left_over = target_size % max_total_character_in_a_line;
	if(left_over)
		memcpy(buffer, target, left_over);

	memcpy(result, original_buffer_pointer, final_size);

	return(final_size);
}

/*	WARNING: to prevent using a temp buffer which size is vary and using malloc()
	I re-use 'source', so 'source' will be modified ... this is also faster ... REMEMBER */
unsigned long PEM_decode(unsigned char *source, long source_size, unsigned char *result)
{
	unsigned long total_lines, left_over;
	unsigned long target_size, final_size;
	unsigned long i;
	unsigned char *target = result;
	unsigned char *source_pointer = source;
	unsigned long *current_value;
	unsigned long total_padding;
	unsigned long total_block;
	unsigned long temp;
	long y; // allow negative value; ======================= IMPORTANT

	// find total_padding
	if(*(source + (source_size - 1)) == pad)
	{
		if(*(source + (source_size - 2)) == pad)
			total_padding = 2;
		else
			total_padding = 1;
	}
	else
		total_padding = 0;

	// find total lines
	total_lines = source_size / max_total_character_in_a_line_plus_CRLF;

	// calculate target size
	target_size = source_size - (total_lines * CRLF_length);

	// calculate final size ... return value 
	final_size = target_size - (target_size / big_size) - total_padding;

	// define total block of 32bits of 6bits
	if(total_padding)
		total_block = (target_size / big_size) - 1;
	else
		total_block = (target_size / big_size);
	
	// get rid of CRLF .. 2 bytes .. store it in target
	for(i = 0; i < total_lines; i++)
	{
		memcpy(target, source_pointer, max_total_character_in_a_line);

		// increase counter
		target += max_total_character_in_a_line;
		source_pointer += max_total_character_in_a_line_plus_CRLF;
	}
	// do left_over if any
	left_over = source_size % max_total_character_in_a_line_plus_CRLF;
	if(left_over)
		memcpy(target, source_pointer, left_over);

	// copy back to the source ... source is modified here
	memcpy(source, result, target_size);

	//clear target .. before we continue ... IS THIS NECESSARY ?
	memset(result, 0, final_size);
		
	// translate 32bits of 6bits to 24bits of 8bits ascii
	source_pointer = source;
	target = result;
	for(i = 0; i < total_block; i++) // do the second block to the end
	{
		current_value = (unsigned long *)target; // direct store in the same variable ... 'result'
		for(y = small_size; y >= 0; y--) // reverse order .. loop 4 times => 3 to 0
		{			
			*current_value <<= six;
			*current_value |= PEM_decode_data[*(source_pointer + y)];
		}

		// increase conter
		source_pointer += big_size;
		target += small_size;
	}
		
	// do the left over if any
	if(total_padding) // if there is padding
	{
		temp = 0;
		for(y = small_size; y >= 0; y--) // reverse order .. loop 4 times => 3 to 0
		{			
			temp <<= six;
			temp |= PEM_decode_data[*(source_pointer + y)];
		}

		memcpy(target, &temp, small_size - total_padding);
	}
	
	return(final_size);
}